# -*- coding: utf-8 -*-
import os
from contextlib import contextmanager

import mlflow
from mlflow.exceptions import MlflowException

from dan import logger


def setup_environment(config: dict):
    """
    Get the necessary variables from the config file and put them in the environment variables

    :param config: dict, the config of the model
    """
    needed_variables = {
        "MLFLOW_S3_ENDPOINT_URL": "s3_endpoint_url",
        "MLFLOW_TRACKING_URI": "tracking_uri",
        "AWS_ACCESS_KEY_ID": "aws_access_key_id",
        "AWS_SECRET_ACCESS_KEY": "aws_secret_access_key",
    }
    for variable_name, config_key in needed_variables.items():
        if config_key in config:
            os.environ[variable_name] = config[config_key]


def logging_metrics(
    display_values: dict,
    step: str,
    epoch: int,
    mlflow_logging: bool = False,
    is_master: bool = False,
):
    """
    Log dictionary metrics in the Metrics section of MLflow

    :param display_values: dict, the dictionary containing the metrics to publish on MLflow
    :param step: str, the step for which the metrics are to be published on Metrics section (ex: train, val, test). This will allow a better display on MLflow.
    :param epoch: int, the current epoch.
    :param mlflow_logging: bool, allows you to verify that you have the authorization to log on MLflow, defaults to False
    :param is_master: bool, makes sure you're on the right thread, defaults to False
    """
    if mlflow_logging and is_master:
        mlflow_values = {
            f"{step}_{name}": value for name, value in display_values.items()
        }
        mlflow.log_metrics(mlflow_values, epoch)


def logging_tags_metrics(
    display_values: dict,
    step: str,
    mlflow_logging: bool = False,
    is_master: bool = False,
):
    """
    Log dictionary metrics in the Tags section of MLflow

    :param display_values: dict, the dictionary containing the metrics to publish on MLflow
    :param step: str, the step for which the metrics are to be published on Tags section (ex: train, val, test). This will allow a better display on MLflow.
    :param mlflow_logging: bool, allows you to verify that you have the authorization to log on MLflow, defaults to False
    :param is_master: bool, makes sure you're on the right thread, defaults to False
    """
    if mlflow_logging and is_master:
        mlflow_values = {
            f"{step}_{name}": value for name, value in display_values.items()
        }
        mlflow.set_tags(mlflow_values)


@contextmanager
def start_mlflow_run(config: dict):
    """
    Create an MLflow execution context with the parameters contained in the config file

    :param config: dict, the config of the model
    """

    # Set needed variables in environment
    setup_environment(config)

    # Set experiment from config
    experiment_id = config.get("experiment_id")
    assert experiment_id, "Missing MLflow experiment ID in the configuration"
    try:
        mlflow.set_experiment(experiment_id=experiment_id)
        logger.info(f"Run Experiment ID : {experiment_id} on MLFlow")
    except MlflowException as e:
        logger.error(f"Couldn't set Mlflow experiment with ID: {experiment_id}")
        raise e

    # Start run
    yield mlflow.start_run(run_name=config.get("run_name"))
    mlflow.end_run()
